# -*- coding: utf-8 -*-
# junky lib: dataset.BertDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementation of torch.utils.data.Dataset for word-level input.
"""
import logging
import sys
#logging.basicConfig(level=logging.ERROR)
#if not sys.warnoptions:
#    import warnings
#    warnings.simplefilter('ignore')
#    os.environ['PYTHONWARNINGS'] = 'ignore'
from junky import CPU, absmax_torch, pad_array_torch, \
                  pad_sequences_with_tensor
from junky.dataset.base_dataset import BaseDataset
from time import time
import torch
from torch import Tensor, tensor
from tqdm import tqdm

# to suppress transformers' warnings
#logging.getLogger('transformers.tokenization_utils_base').setLevel(logging.ERROR)
#logging.getLogger('pytorch_pretrained_bert.tokenization').setLevel(logging.ERROR)
logging.getLogger('transformers').setLevel(logging.ERROR)


class BertDataset(BaseDataset):
    """
    `torch.utils.data.Dataset` for word-level input with contextual
    embeddings.

    Args:
        model: one of the token classification models from the `transformers`
            package. It should be created with config containing
            output_hidden_states=True. NB: Don't forget to set model in the
            eval mode before use it with this class.
        tokenizer: the tokenizer from `transformers` package corresponding to
            `model` chosen.
        int_tensor_dtype: dtype for int tensors: torch.dtype.
        sentences: sequences of words: list([list([str])]). If not ``None``,
            they will be transformed and saved. NB: All the sentences must
            not be empty.
        All other args are params for the `.transform()` method. They are used
            only if *sentences* is not ``None``. You can use any args but
            `save` that is set to `True`.
    """
    overlap_shift = .5
    overlap_border = 2
    use_batch_max_len = True
    sort_dataset = True

    @property
    def vec_size(self):
        return self.data[0].shape[-1] if self.data else \
               self.model.config.hidden_size

    def __init__(self, model, tokenizer, int_tensor_dtype=torch.int64,
                 sentences=None, **kwargs):
        assert model.config.output_hidden_states, \
               'ERROR: BERT model was created with ' \
               '`output_hidden_states=False` option. To use it in ' \
               'BertDataset, the model must be created with ' \
               '`output_hidden_states=True`'
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.int_tensor_dtype = int_tensor_dtype
        if sentences:
            self.transform(sentences, save=True, **kwargs)

    def _pull_xtrn(self):
        xtrn = self.model, self.tokenizer
        self.model, self.tokenizer = None, None
        return xtrn

    def _push_xtrn(self, xtrn):
        self.model, self.tokenizer = xtrn

    @staticmethod
    def _aggregate_hidden_states(hidden_states, layer_ids=None,
                                 aggregate_op='mean'):
        if isinstance(layer_ids, int):
            hiddens = hidden_states[layer_ids]

        else:
            if layer_ids is None:
                hiddens = hidden_states
            else:
                hiddens = [hidden_states[x] for x in layer_ids]

            if aggregate_op == 'absmax':
                hiddens = absmax_torch(hiddens, dim=0)
            elif aggregate_op == 'cat':
                hiddens = torch.cat(hiddens, dim=-1)
            elif aggregate_op == 'first':
                hiddens = hiddens[0]
            elif aggregate_op == 'last':
                hiddens = hiddens[-1]
            elif aggregate_op == 'max':
                hiddens = torch.max(hiddens, dim=0)[0]
            elif aggregate_op == 'mean':
                hiddens = torch.mean(hiddens if isinstance(hiddens, Tensor) else
                                     torch.stack(hiddens, dim=0),
                                     dim=0)
            elif aggregate_op == 'sum':
                hiddens = torch.sum(hiddens if isinstance(hiddens, Tensor) else
                                     torch.stack(hiddens, dim=0),
                                     dim=0)
            else:
                RuntimeError(
                    'ERROR: unknown aggregate_op '
                    "(choose one of ['absmax', 'cat', 'max', 'mean', 'sum'])"
                )
        return hiddens

    def transform(self, sentences, max_len=None, batch_size=64,
                  hidden_ids=0, aggregate_hiddens_op='mean',
                  aggregate_subtokens_op='absmax', to=None,
                  save=True, append=False, loglevel=1):
        """Convert *sentences* of words to the sequences of the corresponding
        contextual vectors and adjust their format for Dataset.

        *max_len* is a param for tokenizer. We'll transform lines of any
            length, but the quality is higher if *max_len* is greater.
            ``None`` (default) or `0` means the maximum for the model
            (usually, `512`).

        *batch_size* affects only on the execution time. Greater is faster,
            but big *batch_size* may be cause of CUDA Memory Error. If
            ``None`` or `0`, we'll try to convert all *sentences* with one
            batch.

        *hidden_ids*: hidden score layers that we need to aggregate. Allowed
            int or tuple of ints. If ``None``, we'll aggregate all the layers.

        *aggregate_hidden_op*: how to aggregate hidden scores. The ops
            allowed: 'absmax', 'cat', 'max', 'mean', 'sum'. For the 'absmax'
            method we take into account absolute values of the compared items.

        *aggregate_subtokens_op*: how to aggregate subtokens vectors to form
            only one vector for each input token. The ops allowed: ``None``,
            'absmax', 'first', 'last', 'max', 'mean', 'sum'. For the 'absmax'
            method we take into account absolute values of the compared items.

        If you want to get the result placed on some exact device, specify the
        device with *to* param. If *to* is ``None`` (defautl), data will be
        placed to the very device that `bs.model` is used.

        If *save* is ``True``, we'll keep the converted sentences as the
        `Dataset` source.

        If *append* is ``True``, we'll append the converted sentences to the
        existing Dataset source. Elsewise (default), the existing Dataset
        source will be replaced. The param is used only if *save* is ``True``.

        *loglevel* can be set to `0`, `1` or `2`. `0` means no output.

        The result is depend on *aggregate_subtokens_op* param. If it is
        ``None``, then for each token we keeps in the result a tensor with
        stacked vectors for all its subtokens. Otherwise, if any
        *aggregate_subtokens_op* is used, each sentence will be converted to
        exactly one tensor of shape [<sentence length>, <vector size>]."""

        tokenizer_max_len = self.tokenizer.max_len \
                                if hasattr(self.tokenizer, 'max_len') else \
                            self.tokenizer.model_max_len \
                                if hasattr(self.tokenizer,
                                           'model_max_len') else \
                            512

        if not max_len:
            max_len = tokenizer_max_len
        assert max_len >= 16, 'ERROR: max len must be >= 16'
        assert max_len <= tokenizer_max_len, \
               'ERROR: max len must be <= {}'.format(tokenizer_max_len)
        valid_ops = ['absmax', 'cat', 'max', 'mean', 'sum']
        assert aggregate_hiddens_op in valid_ops, \
               'ERROR: unknown aggregate_hidden_op (choose one of {})' \
                   .format(valid_ops)
        valid_ops = [None, 'absmax', 'first', 'last', 'max', 'mean', 'sum']
        assert aggregate_subtokens_op in valid_ops, \
               'ERROR: unknown aggregate_subtokens_op (choose one of {})' \
                   .format(valid_ops)
        if self.data and append:
            assert (
                isinstance(self.data[0], list)
            and aggregate_subtokens_op is None
            ) or (
                isinstance(self.data[0], Tensor)
            and aggregate_subtokens_op is not None
            ), "ERROR: can't append data created with inconsistent " \
               'aggregate_subtokens_op'
        device = next(self.model.parameters()).device
        max_len_ = max_len - 2  # for [CLS] and [SEP]

        overlap_border = max_len if self.overlap_border is None else \
                         0 if self.overlap_border < 0 else \
                         self.overlap_border
        shift = int(max(min(max_len_, self.overlap_shift)
                        if self.overlap_shift >= 1 else
                    (max_len_ * self.overlap_shift), 1))

        _src = sentences
        if loglevel >= 2:
            print('Tokenizing')
            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
        time0 = time()
        # tokenize each token separately by the BERT tokenizer
        tokenized_sentences = \
            [[x for x in self.tokenizer.tokenize(x) or ['[UNK]']]
                 for x in _src] \
                if sentences and isinstance(sentences[0], str) else \
            [[self.tokenizer.tokenize(x) or ['[UNK]'] for x in x] \
                 for x in _src]
        if time() - time0 < 2 and loglevel == 1:
            loglevel = 0

        # number of subtokens in tokens of tokenized_sentences
        num_subtokens = [[len(x) for x in x] for x in tokenized_sentences]
        # for each subtoken we keep index of its token
        sub_to_kens = [[
            x for x in [[i] * x for i, x in enumerate(x)]
              for x in x
        ] for x in num_subtokens]
        # for each token we keep its start in flatten tokenized_sentences
        token_starts = []
        for sent in sub_to_kens:
            starts, prev_idx = [], -1
            for i, idx in enumerate(sent):
                if idx != prev_idx:
                    starts.append(i)
                    prev_idx = idx
            token_starts.append(starts)

        # flattening tokenized_sentences
        tokenized_sentences = [[x for x in x for x in x]
                                   for x in tokenized_sentences]
        # lengths of flattened tokenized_sentences
        sent_lens = [len(x) for x in tokenized_sentences]

        need_sort_dataset = self.sort_dataset \
                        and self.use_batch_max_len \
                        and len(tokenized_sentences) > batch_size
        ## sort tokenized sentences by lenght
        ######
        if need_sort_dataset:
            sent_lens, sorted_sent_ids = zip(*sorted(
                [(x, i) for i, x in enumerate(sent_lens)],
                reverse=True
            ))
            sent_lens = list(sent_lens)
            tokenized_sentences = [
                tokenized_sentences[i] for i in sorted_sent_ids
            ]
            sub_to_kens = [
                sub_to_kens[i] for i in sorted_sent_ids
            ]
            token_starts = [
                token_starts[i] for i in sorted_sent_ids
            ]
            ######

        def process_long_sentences(sents, sent_lens, sub_to_kens,
                                   token_starts):
            overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
            overlap_token_starts, overmap = [], [], [], [], []
            for i, (sent, sent_len, token_ids, sub_ids) \
                    in enumerate(zip(sents, sent_lens, sub_to_kens,
                                     token_starts)):
                if sent_len > max_len_:
                    # находим индекс токена для сабтокена с шифтом
                    pos = token_ids[shift]
                    # вычитаем токен нулевого сабтокена, получаем индекс
                    # токена относительно текущего начала
                    pos_ = pos - token_ids[0]
                    if not pos_:
                        pos, pos_ = token_ids[sub_ids[1] - sub_ids[0]], 1
                    # находим индекс сабтокена: из индекса сабтокена
                    # найденного токена вычитаем индекс текущего нулевого
                    # сабтокена
                    start = sub_ids[pos_] - sub_ids[0]
                    if start > max_len_:
                        raise RuntimeError(
                            ('ERROR: too long token (longer than '
                             'effective max_len):\n{}')
                                 .format(sent[:start])
                        )
                    overlap_sents.append(sent[start:])
                    overlap_sent_lens.append(sent_len - start)
                    overlap_sub_to_kens.append(token_ids[start:])
                    overlap_token_starts.append(sub_ids[pos_:])
                    overmap.append((i, pos_))
                    end = token_ids[max_len_] - token_ids[0]
                    end = sub_ids[end] - sub_ids[0]
                    sent[end:] = []
            return overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
                   overlap_token_starts, overmap

        overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
        overlap_token_starts, overmap = process_long_sentences(
            tokenized_sentences, sent_lens,
            sub_to_kens, token_starts
        )
        num_sents = len(tokenized_sentences)
        overmap = {i + num_sents: (orig_i, pos)
                       for i, (orig_i, pos) in enumerate(overmap)}

        while overlap_sents:
            tokenized_sentences += overlap_sents
            sent_lens += overlap_sent_lens
            sub_to_kens += overlap_sub_to_kens
            token_starts += overlap_token_starts
            prev_num_sents = num_sents
            num_sents = len(tokenized_sentences)
            overlap_sents, overlap_sent_lens, overlap_sub_to_kens, \
            overlap_token_starts, overmap_ = process_long_sentences(
                overlap_sents, overlap_sent_lens,
                overlap_sub_to_kens, overlap_token_starts
            )
            for i, (orig_i, pos) in enumerate(overmap_):
                prev_map = overmap[orig_i + prev_num_sents]
                overmap[i + num_sents] = (prev_map[0], prev_map[1] + pos)

        splitted_sent_lens = [len(x) for x in tokenized_sentences]

        if not batch_size:
            batch_size = num_sents

        data = []
        _src = range(0, num_sents, batch_size)
        if loglevel:
            print('Vectorizing')
            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)

        for batch_i in _src:
            batch_max_len = min(
                max(splitted_sent_lens[batch_i:batch_i + batch_size]) + 2,
                max_len
            ) if self.use_batch_max_len else max_len

            encoded_sentences = [
                self.tokenizer.encode_plus(text=sent,
                                           add_special_tokens=True,
                                           max_length=batch_max_len,
                                           pad_to_max_length=True,
                                           #truncation=True,
                                           return_tensors='pt',
                                           return_attention_mask=True,
                                           return_overflowing_tokens=False)
                    for sent in tokenized_sentences[batch_i:batch_i
                                                  + batch_size]
            ]
            input_ids, attention_masks = zip(*[
                (x['input_ids'], x['attention_mask'])
                    for x in encoded_sentences
            ])

            with torch.no_grad():
                hiddens = self.model(
                    torch.cat(input_ids, dim=0).to(device),
                    token_type_ids=None,
                    attention_mask=torch.cat(attention_masks, dim=0)
                                        .to(device)
                )[-1]

                hiddens = self._aggregate_hidden_states(
                    hiddens, layer_ids=hidden_ids,
                    aggregate_op=aggregate_hiddens_op
                )

                if to:
                    hiddens = hiddens.to(to)

                for i, sent in enumerate(hiddens, start=batch_i):
                    sent_len = splitted_sent_lens[i]
                    if i in overmap:
                        j, over_pos_start = overmap[i]
                        over_pos_end = sub_to_kens[j][len(data[j])]
                        overlap = over_pos_end - over_pos_start
                        if overlap > overlap_border * 2:
                            start1 = over_pos_start + overlap_border
                            end1 = over_pos_end - overlap_border
                            overlap = end1 - start1
                            half2 = overlap // 2
                            half1 = overlap - half2
                            half = half1 + (1 if half1 == half2 else 0)
                            for k in range(half2):
                                coef = (k + 1) / half / 2
                                data[j][start1 + k] = \
                                    data[j][start1 + k] * (1 - coef) \
                                  + sent[overlap_border + k] * coef
                                k_ = overlap - k - k
                                data[j][start1 + k_] = \
                                    data[j][start1 + k_] * coef \
                                  + sent[overlap_border + k_] * (1 - coef)
                            if half1 != half2:
                                data[j][start1 + half1] = (
                                    data[j][start1 + half1]
                                  + sent[overlap_border + half1]
                                ) / 2
                            end1 = token_starts[j][end1]
                        else:
                            start2 = overlap - overlap // 2
                            end1 = token_starts[j][over_pos_start + start2]

                        start2 = end1 - token_starts[j][over_pos_start]
                        data[j] = torch.cat([
                            data[j][:end1], sent[1 + start2:1 + sent_len]
                        ], dim=0)
                    else:
                        data.append(sent[1:1 + sent_len])

        ## sort data in original order
        ######
        if need_sort_dataset:
            _, data = zip(*sorted(
                [(i, x) for i, x in enumerate(data)],
                key=lambda x: sorted_sent_ids[x[0]]
            ))
            data = list(data)
        ######

        _src = num_subtokens
        if loglevel:
            print('Reordering')
            _src = tqdm(iterable=_src, mininterval=2, file=sys.stdout)
        for i, token_lens in enumerate(_src):
            token_lens = num_subtokens[i]
            start = 0
            sent = data[i]
            sent_ = []
            for token_len in token_lens:
                end = start + token_len
                vecs = sent[start:end]
                if aggregate_subtokens_op != None:
                    vecs = self._aggregate_hidden_states(
                        vecs, layer_ids=None,
                        aggregate_op=aggregate_subtokens_op
                    )
                sent_.append(vecs)
                start = end

            data[i] = sent_ if aggregate_subtokens_op == None else \
                      torch.stack(sent_, dim=0)

        if save:
            if append:
                self.data += data
            else:
                self.data = data
        else:
            return data

    def _collate(self, batch, with_lens=True, with_token_lens=True):
        """The method to use with `torch.utils.data.DataLoader` and
        `.transform_collate()`.

        :with_lens: return lengths of data.
        :with_token_lens: return lengths of tokens of the data.
        :return: depends on keyword args.
        :rtype: if the `.transform()` method was called with
            *aggregate_subtokens_op*=None:
                tuple(list([torch.tensor]), lens:torch.tensor,
                      token_lens:list([torch.tensor]))
            otherwise: tuple(list([torch.tensor]), lens:torch.tensor)
        """
        device = CPU
        pad = 0.

        if isinstance(batch[0], Tensor):
            if batch[0].is_cuda:
                device = batch[0].get_device()
            lens = [tensor([len(x) for x in batch], device=device,
                           dtype=self.int_tensor_dtype)] if with_lens else []
            x = pad_sequences_with_tensor(batch, padding_tensor=pad)

        else:
            for x in batch:
                if x:
                    if x[0].is_cuda:
                        device = x[0].get_device()
                    tensor_dtype = x[0].dtype
                    break
            lens = [tensor([len(x) for x in batch], device=device,
                           dtype=self.int_tensor_dtype)] if with_lens else []
            if with_token_lens:
                lens.append([tensor([len(x) for x in x], device=device,
                                    dtype=self.int_tensor_dtype)
                                 for x in batch])
            x = pad_array_torch(batch, padding_value=pad,
                                device=device, dtype=tensor_dtype)

        return (x, *lens)



# -*- coding: utf-8 -*-
# junky lib: plotter
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a bunch of tools to visualize train results.
"""
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix

import matplotlib as mpl
import matplotlib.style
from matplotlib import pyplot as plt
import seaborn as sns


def plot_losses(train_losses=None, test_losses=None, accuracies=None,
                plot_title='Train/Dev Loss', figsize=(7, 4), legend_loc='best',
                legend_labels=['train', 'dev', '1 - acc'], save_name=None):
    """Plots train and dev losses obtained during training.
    The plot image is saved to disk.
    
    Args:
    
      train_losses:   list of floats, train losses throughout epochs.
                      If `None`, not plotted.
      test_losses:    list of floats, test losses throughout epochs.
                      If `None`, not plotted.
      accuracies:     optional, list of floats, accuracies throughout epochs.
                      Here, used to count `(1 - accuracy)`.
                      If `None`, not plotted.
      plot_title:     plot title, `str`. Default value - 'Train/Dev Loss'.
      figsize:        the size of the figure plotted. Default size is `(5,3)`.
      legend_loc:     location of the legend according to matplotlib.
                      Default: 'best'. If invalid `legend_loc` is entered,
                      matplotlib will place the legend in the `best` location and
                      show a warning with all possible valid loc definitions.
      legend_labels:  set legend labels. Default: `['train', 'dev', '1 - acc']`
      save_name:      optional, if `None`, plot is not saved.
                      Used as `fname` in `plt.savefig()`.
                      Default file extention is '.png', if other extention is needed,
                      please specify extention in save_name as well.
                      Example: save_name='plot.pdf'
    """
    mpl.style.use('default')
    plt.figure(figsize=figsize)

    if train_losses:
        plt.plot([None] + train_losses)
    if test_losses:
        plt.plot([None] + test_losses)
    if accuracies is not None:
        plt.plot([None] + [1 - x for x in accuracies])
    plt.title(plot_title)
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend(legend_labels, loc=legend_loc)
    plt.grid()
    if save_name is not None:
        plt.savefig(save_name, bbox_inches='tight')
    plt.show()

def plot_confusion_matrix(y_true, y_pred,
                          pad_index=None, ymap=None, figsize=(20, 10),
                          show_total=['x', 'y'], show_zeros=True, show_empty_tags=False,
                          plot_title='Confusion Matrix', save_name=None):
    """
    Generate matrix plot of confusion matrix with pretty annotations.
    The plot image is saved to disk.
    
    Args:
    
      y_true:           true labels of the data, with shape (nsamples,)

      y_pred:           label predictions, with shape (nsamples,)

      pad_index:        if not `None` and not present in `y_pred`, `pad_index`
                        will not be included in the plot.

      ymap:             dict: index -> tag.
                        if not `None`, map the predictions to their categorical labels.
                        if `None`, `range(1, len(set(y_true+y_pred))`
                        is used for labels.

      figsize:          tuple: the size of the figure plotted.

      show_total:       list of `str`. Where to display total number of
                        class occurrences in the corpus: diagonal and/or axes.
                        Up to all from `['diag', 'x', 'y']` can be chosen.
                        Default = `['x', 'y']` (horizontal and vertical axes respectively).
                        If `None`, total values are not displayed on the plot.

      show_zeros:       bool: whether to show zeros in the confusion matrix.

      show_empty_tags:  only active when when `ymap` is specified.
                        If `True`, all tags, including those that weren't met
                        neither in `y_true` or `y_pred` are displayed
                        (filled with `0` in both axes).
                        NB! If `True`, `pad_idx` will also be displayed even if specified.
                        Default: `False`, 'empty' tags are skipped.

      plot_title:       str: plot title, default title - 'Confusion Matrix'.

      save_name:        str: filename of figure file to save.
                        if `None`, image is not saved to disk.
    """
    # if pad_index is not None and does not exist in y_pred, it's excluded
    # from labels
    mpl.style.use('default')

    cm = confusion_matrix(y_true, y_pred)

    if ymap:
        if show_empty_tags:
            for i in ymap.keys():
                if i not in set(y_true+y_pred):
                    cm = np.insert(cm, i, 0, axis=0)
                    cm = np.insert(cm, i, 0, axis=1)
            labels = ymap.values()

        if not show_empty_tags:

            labels = [ymap[i] if pad_index is not None and i != pad_index
                                         and pad_index not in y_pred else
                      ymap[i]
                      for i in set(y_true+y_pred)]

    else:
        labels = [i if pad_index is not None and i != pad_index
                                         and pad_index not in y_pred else
              i
              for i in set(y_true+y_pred)]

    cm_sum = np.sum(cm, axis=1, keepdims=True)
    cm_perc = np.divide(cm, cm_sum.astype(float), where=cm_sum!=0)*100


    annot = np.empty_like(cm).astype(str)
    nrows, ncols = cm.shape
    for i in range(nrows):
        for j in range(ncols):
            c = cm[i, j]
            p = cm_perc[i, j]
            if i == j:
                if show_total and 'diag' in show_total:
                    s = cm_sum[i]
                    annot[i, j] = '%.2f%%\n%d/%d' % (p, c, s)
                else:
                    annot[i, j] = '%.2f%%\n%d' % (p, c)
            elif c == 0:
                if show_zeros:
                    annot[i, j] = '0'
                else:
                    annot[i, j] = ''
            else:
                annot[i, j] = '%.2f%%\n%d' % (p, c)

    total_labels = [str(i)+'\n'+str(n[0]) for i, n in zip(labels, cm_sum)]

    cm = pd.DataFrame(cm,
                      index=total_labels if show_total and 'y' in show_total else labels,
                      columns=total_labels if show_total and 'x' in show_total else labels)
    cm.index.name = 'Actual'
    cm.columns.name = 'Predicted'

    fig, ax = plt.subplots(figsize=figsize)
    sns.heatmap(cm, annot=annot, fmt='', ax=ax)
    plt.title(plot_title)

    if save_name:
        plt.savefig(save_name, bbox_inches='tight')

    plt.show()

def plot_metrics(metrics=[], legend_loc='best',
                 labels=['accuracy', 'precision', 'recalls', 'f1_score'],
                 plot_title='Metrics', figsize=(7, 4), save_name=None):
    """Plots metrics obtained during training.
    Default: ['accuracy', 'precision', 'recalls', 'f1_score'].
    The plot image is saved to disk.
    
    Args:
    
      metrics:        tuple or list of metrics, where each metric is
                      a list of floats, len(metric)==num_epochs.
      legend_loc:     location of the legend according to matplotlib.
                      Default: 'best'. If invalid `legend_loc` is entered,
                      matplotlib will place the legend in the `best` location and
                      show a warning with all possible valid loc definitions.
      labels:         list of str, labels for metrics plotted.
      figsize:        tuple: the size of the figure plotted.
      plot_title:     str: plot title, default title - 'Metrics'.
      save_name:      str: filename of figure file to save.
                      if `None`, image is not saved to disk.
    """
    mpl.style.use('default')
    plt.figure(figsize=figsize)
    for metric in metrics:
        plt.plot([None] + metric)
    plt.grid()
    plt.title(plot_title)
    plt.xlabel('Epochs')
    plt.ylabel('Score')
    plt.legend(labels, loc=legend_loc)
    if save_name is not None:
        plt.savefig(save_name, bbox_inches='tight')
    plt.show()

def plot_seqlens(corpus, bins=20, figsize=(7, 2), **plot_kwargs):
    """Plots histogram of sentence lengths from the corpus.

    Args:

    **corpus** (`list([list([str])])`): pretokenized corpus.

    **figsize** (`tuple(x, y)`): figure size.

    **\*\*plot_kwargs**: any other `pandas.DataFrame.hist` 
    keyword arguments.
    """
    mpl.style.use('default')
    lens = [len(x) for x in corpus]
    _ = pd.Series(lens).hist(bins=bins, figsize=figsize, **plot_kwargs)
    print('Minimum sequence length:', min(lens))
    print('Maximum sequence length:', max(lens))



# -*- coding: utf-8 -*-
# junky lib: dataset.LabelDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementation of torch.utils.data.Dataset for label-level input.
"""
from junky import CPU, make_token_dict
from junky.dataset.base_dataset import BaseDataset
from torch import Tensor, int64, stack, tensor, zeros


class LabelDataset(BaseDataset):
    """
    torch.utils.data.Dataset for label-level input.

    Args:
        labels: list([str])|list[list([str])]).
        unk_label: add a value for labels that are not present in the internal
            dict: str.
        extra_labels: add label values for any other purposes: list([str]).
        tensor_dtype: dtype for tensors' data: torch.dtype. Default is torch.int64
        transform: if `True`, transform and save `labels`.
        skip_unk, keep_empty: params for the `.transform()` method.
    """
    def __init__(self, labels, unk_label=None, extra_labels=None, 
                 tensor_dtype=int64, transform=False, skip_unk=False,
                 keep_empty=False):
        super().__init__()
        self.tensor_dtype = tensor_dtype
        self.fit(labels, unk_label=unk_label, extra_labels=extra_labels)
        if transform:
            self.transform(labels, skip_unk=skip_unk, keep_empty=keep_empty,
                           save=True)

    def fit(self, labels, unk_label=None, extra_labels=None):
        """Recreate the internal dict.

        :param labels: sequences of label values.
        :type labels: list([str])|list[list([str])]).
        :param unk_label: add a value for labels that are not present in the
            dict.
        :type unk_label: str
        :param extra_labels: add label values for any other purposes.
        :type extra_labels: list([str])
        """
        if unk_label:
            extra_labels = (extra_labels if extra_labels else []) \
                         + [unk_label]
        self.transform_dict, _, extra = make_token_dict(
             labels if labels and (isinstance(labels[0], list)
                                or isinstance(labels[0], tuple)) else
             [labels],
             extra_tokens=extra_labels
        )
        self.unk = extra[-1] if unk_label else None
        self.reconstruct_dict = {y: x for x, y in self.transform_dict.items()}

    def label_to_idx(self, label, skip_unk=False):
        """Convert a label value to its index. If the value is not present in
        the internal dict, return index of unk label or None if it's not
        defined."""
        return self.transform_dict[label] \
                   if label in self.transform_dict else \
               self.unk if not skip_unk and self.unk is not None else \
               None

    def idx_to_label(self, idx, skip_unk=False):
        """Convert an index to the corresponding label value. If the index is
        not present in the internal dict, return unk label or empty string if
        it's not defined or *skip_unk* is `True`."""
        if isinstance(idx, Tensor):
            idx = idx.tolist()
        return self.reconstruct_dict[idx] \
                   if idx in self.reconstruct_dict else \
               '' if skip_unk or self.unk is None else \
               self.reconstruct_dict[self.unk]

    def transform(self, labels, skip_unk=False, keep_empty=False,
                  save=True, append=False):
        """Convert *labels* of str type to the sequences of the corresponding
        indices and adjust their format for Dataset. If *skip_unk* is
        `True`, unknown labels will be skipped. If *keep_empty* is `False`,
        we'll remove rows that have no data after converting.

        Corresponding indices are represented as is (int numbers) if *labels*
        are of list([str]) type. Elsewise, if *labels* are of list of
        list([str]) type (each row may contain several label values), the
        indices are represented as multi-hot vectors.

        The type of indices representation is exactly the tensor_dtype
        specified in constructor.

        If *save* is `True`, we'll keep the converted labels as the
        Dataset source.

        If *append* is `True`, we'll append the converted labels to the
        existing Dataset source. Elsewise (default), the existing Dataset
        source will be replaced. The param is used only if *save* is
        `True`."""

        if labels and (isinstance(labels[0], list)
                    or isinstance(labels[0], tuple)):
            data = []
            for labs in labels:
                d = zeros((len(self.transform_dict),),
                          dtype=self.tensor_dtype)
                for i in (self.label_to_idx(l, skip_unk=skip_unk)
                              for l in labs):
                    d[i] = 1
                data.append(d)
        else:
            data = [tensor(i, dtype=self.tensor_dtype)
                        for i in (self.label_to_idx(l, skip_unk=skip_unk)
                                      for l in labels)
                        if keep_empty or i is not None]
        if save:
            if append:
                self.data += data
            else:
                self.data = data
        else:
            return data

    def reconstruct(self, ids, skip_unk=False, keep_empty=False):
        """Convert *sequences* of indices in Dataset format to the rows of the
        corresponding label values.
        If *skip_unk* is `True`, unknown indices will be skipped.
        If *keep_empty* is `False`, we'll remove labels that are empty after
        converting."""
        return [l for l in (self.idx_to_label(i, skip_unk=skip_unk)
                                for i in ids)
                    if keep_empty or l]

    def fit_transform(self, labels, unk_label=None, extra_labels=None,
                      skip_unk=False, keep_empty=False, save=True):
        """Just a serial execution `.fit()` and `.transform()` methods."""
        self.fit(labels, unk_label=unk_label, extra_labels=extra_labels)
        return self.transform(labels, skip_unk=skip_unk,
                              keep_empty=keep_empty, save=save)

    def _collate(self, batch):
        #return tensor(batch, dtype=self.tensor_dtype)
        return stack(batch, dim=0)



# -*- coding: utf-8 -*-
# junky lib: layers.Masking
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a Masking layer implementation for PyTorch models.
"""
from collections.abc import Iterable
from junky import CPU
import torch
import torch.nn as nn


class Masking(nn.Module):
    """
    Replaces certain elemens of the incoming data to the `mask` given.

    Args:
        input_size: The number of expected features in the input `x`.
        mask: Replace to what.
        indices_to_highlight: What positions in the `feature` dimension of the
            masked positions of the incoming data must not be replaced to the
            `mask`.
        highlighting_mask: Replace data in that positions to what. If
            ``None``, the data will keep as is.
        batch_first: If ``True``, then the input and output tensors are
            provided as `(batch, seq, feature)` (<==> `(N, *, H)`). Default:
            ``False``.

    Shape:
        - Input: :math:`(*, N, H)` where :math:`*` means any number of
          additional dimensions and :math:`H = \text{input_size}`.
        - Output: :math:`(*, N, H)` where all are the same shape as the input
          and :math:`H = \text{input_size}`.

    .. note:: Masking layer was made for using right before Softmax. In that
        case and with `mask`=``-inf`` (default), the Softmax output will have
        zeroes in all positions corresponding to `indices_to_mask`.

    .. note:: Usually, you'll mask positions of all non-pad tags in padded
        endings of the input data. Thus, after Softmax, you'll always have the
        padding tag predicted for that endings. As the result, you'll have
        loss = 0, that prevents your model for learning on padding.

    Examples::

        >>> m = Masking(4, batch_first=True)
        >>> input = torch.randn(2, 3, 4)
        >>> output = m(input, torch.tensor([1, 3]))
        >>> print(output)
        tensor([[[ 1.1912, -0.6164,  0.5299, -0.6446],
                 [   -inf,    -inf,    -inf,  1.0000],
                 [   -inf,    -inf,    -inf,  1.0000]],

                [[-0.3011, -0.7185,  0.6882, -0.1656],
                 [-0.3316, -0.3521, -0.9717,  0.5551],
                 [ 0.7721,  0.2061,  0.8932, -1.5827]]])
    """
    __constants__ = ['batch_first', 'highlighting_mask',
                     'indices_to_highlight', 'input_size', 'mask']

    def __init__(self, input_size, mask=float('-inf'),
                 indices_to_highlight=-1, highlighting_mask=1,
                 batch_first=False):
        super().__init__()

        if not isinstance(indices_to_highlight, Iterable):
            indices_to_highlight = [indices_to_highlight]

        self.input_size = input_size
        self.mask = mask
        self.indices_to_highlight = indices_to_highlight
        self.highlighting_mask = highlighting_mask
        self.batch_first = batch_first

        output_mask = torch.tensor([mask] * input_size)
        if indices_to_highlight is not None:
            if highlighting_mask is None:
                output_mask0 = torch.tensor([0] * input_size,
                                            dtype=output_mask.dtype)
                for idx in indices_to_highlight:
                    output_mask0[idx] = 1
                    output_mask[idx] = 0
                output_mask = torch.stack((output_mask0, output_mask))
            else:
                for idx in indices_to_highlight:
                    output_mask[idx] = highlighting_mask
        self.register_buffer('output_mask', output_mask)

    def forward(self, x, lens):
        """
        :param lens: array of lengths of **x** by the `seq` dimension.
        """
        output_mask = self.output_mask
        output_mask0, output_mask = \
            output_mask if len(output_mask.shape) == 2 else \
            (None, output_mask)
        device = output_mask.get_device() if output_mask.is_cuda else CPU
        if not isinstance(lens, torch.Tensor):
            lens = torch.tensor(lens, device=device)

        seq_len = x.shape[self.batch_first]
        padding_mask = \
            torch.arange(seq_len, device=device) \
                 .expand(lens.shape[0], seq_len) >= lens.unsqueeze(1)
        if not self.batch_first:
            padding_mask = padding_mask.transpose(0, 1)
        x[padding_mask] = output_mask if output_mask0 is None else \
                          x[padding_mask] * output_mask0 + output_mask

        return x

    def extra_repr(self):
        return ('{}, mask={}, indices_to_highlight={}, highlighting_mask={}, '
                'batch_first={}').format(
                    self.input_size, self.mask, self.indices_to_highlight,
                    self.highlighting_mask, self.batch_first
                )



# -*- coding: utf-8 -*-
# junky lib: dataset.FrameDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
A frame for use several torch.utils.data.Dataset together.
"""
from collections import OrderedDict
from junky.dataset.base_dataset import BaseDataset


class FrameDataset(BaseDataset):
    """
    A frame for use several objects of `junky.dataset.Dataset` conjointly.
    All the datasets must have the data of equal length.
    """
    def __init__(self):
        super().__init__()
        delattr(self, 'data')
        self.datasets = OrderedDict()

    def __len__(self):
        """Returns the lentgh of the first added dataset. Note, that all
        datasets must be of equal length."""
        return len(next(iter(self.datasets.values()))[0]) \
                   if self.datasets else \
               0

    def __getitem__(self, idx):
        """Returns a tuple of outputs of all added datasets in order of
        addition."""
        return tuple(x for x in self.datasets.values()
                       for x in (x[0][idx]
                                     if isinstance(x[0][idx], tuple) else
                                 [x[0][idx]]))

    def _pull_data(self):
        data = {}
        for name, ds in self.datasets.items():
            data[name] = ds[0]._pull_data()
        return data

    def _push_data(self, data):
        for name, d in data.items():
            self.datasets[name][0]._push_data(d)

    def _pull_xtrn(self):
        xtrn = {}
        for name, ds in self.datasets.items():
            xtrn[name] = ds[0]._pull_xtrn()
        return xtrn

    def _push_xtrn(self, xtrn):
        for name, x in xtrn.items():
            self.datasets[name][0]._push_xtrn(x)

    def add(self, name, dataset, **collate_kwargs):
        """Add *dataset* with a specified *name*.

        :param **collate_kwargs: keyword arguments for the *dataset*'s
            `._frame_collate()` method.
        """
        assert name not in self.datasets, \
               "ERROR: dataset '{}' was already added".format(name)
        assert len(dataset) > 0, "ERROR: can't add empty dataset"
        num_pos = len(dataset[0]) if isinstance(dataset[0], tuple) else 1
        self.datasets[name] = dataset, num_pos, collate_kwargs

    def remove(self, name):
        """Remove dataset with a specified *name*."""
        del self.datasets[name]

    def get(self, name):
        """Get dataset with a specified *name*.

        :return: dataset, collate_kwargs
        """
        return self.datasets[name][0], self.datasets[name][2]

    def get_dataset(self, name):
        """Get dataset with a specified *name*.

        :return: dataset
        """
        return self.datasets[name][0]

    def list(self):
        """Get names of the nested datasets in order of their addition."""
        return tuple(self.datasets.keys())

    def to(self, *args, **kwargs):
        """Invoke the `.to()` method for all object of `torch.Tensor` or
        `torch.nn.Module` type."""
        [x[0].to(*args, **kwargs) for x in self.datasets.values()]

    def transform(self, sentences, names=None, save=True, append=False,
                  part_kwargs=None, **kwargs):
        """Invoke `.transform()` methods for nested `Dataset` objects.

        *names* is a list of datasets `.transform()` methods of which will be
        called.

        *save*, *append* and **kwargs will be transfered to any nested
        `.transform()` methods.

        *part_kwargs* is a `dict` of format: {<name>: kwargs, ...}, where one
        can specify separate keyword args for `.transform()` metod of certain
        nested `Dataset` objects.

        If *save* is ``False``, we'll return the stacked result of objects'
        returns."""
        if isinstance(names, str):
            names = [names]
        data = tuple(
            y[0].transform(sentences, save=save, append=append, **kwargs,
                           **(part_kwargs[x]
                                  if part_kwargs and x in part_kwargs else
                              {}))
                for x, y in self.datasets.items()
                    if names is None
                    or x in names
        )
        if not save:
            return data

    def _frame_collate(self, batch, pos, names=None):
        """The method to use with junky.dataset.FrameDataset.

        :param pos: position of the data in *batch*.
        :type pos: int
        """
        if isinstance(names, str):
            names = [names]
        res = []
        for name, (ds, num_pos, kwargs) in self.datasets.items():
            if names is None or name in names:
                res_ = ds._frame_collate(batch, pos, **kwargs)
                res += res_ if isinstance(res_, tuple) else [res_]
                pos += num_pos
        return tuple(res)

    def _collate(self, batch, names=None):
        """The method to use with `torch.utils.data.DataLoader` and
        `.transform_collate()`. It concatenates outputs of the added datasets
        in order of addition. All the dataset must have the method
        `.frame_collate(batch, pos)`, where *pos* is the first position of the
        corresponding dataset's data in the batch."""
        return self._frame_collate(batch, 0, names=names)



# -*- coding: utf-8 -*-
# junky lib: dataset.BertDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementation of torch.utils.data.Dataset that prepare text
sentences for `transformers.BertModel` input.
"""
from junky.dataset.base_dataset import BaseDataset
import logging
import sys
from torch import int64, tensor
from tqdm import tqdm

logging.getLogger('transformers').setLevel(logging.ERROR)


class BertTokenizedDataset(BaseDataset):
    """
    `torch.utils.data.Dataset` for `transformers.BertModel` input.

    Args:
        tokenizer: the tokenizer from `transformers` package corresponding to
            `model` chosen.
        int_tensor_dtype: dtype for int tensors: torch.dtype.
        sentences (`list([str])|list([list([str])])`): If not ``None``, they
            will be transformed and saved. NB: All the sentences must not be
            empty.
        All other args are params for the `.transform()` method. They are used
            only if *sentences* is not ``None``. You can use any args but
            `save` that is set to `True`.
    """
    def __init__(self, tokenizer, int_tensor_dtype=int64,
                 sentences=None, **kwargs):
        super().__init__()
        self.tokenizer = tokenizer
        self.int_tensor_dtype = int_tensor_dtype
        if sentences:
            self.transform(sentences, save=True, **kwargs)

    def _pull_xtrn(self):
        xtrn = self.tokenizer
        self.tokenizer = None
        return xtrn

    def _push_xtrn(self, xtrn):
        self.tokenizer = xtrn

    def transform(self, sentences, add_special_tokens=True, is_pretokenized=False,
                  max_len=None, save=True, append=False):
        """Convert text *sentences* to the `transformers.BertModel` input.
        Already tokenized sentences are also allowed but fill be joined before
        tokenizing with space character.

        *max_len*, *add_special_tokens* and *is_pretokenized* are params for
            the tokenizer. *max_len* `0` means the highest number of subtokens
            for the model (usually, `512`). ``None`` (default) means no limit.

        Important! `is_pretokenized=True` means, that each sentence is
        splitten on tokens ("pre-tokenized"), not tokenized with some
        transformers' tokenizer. If sentences are fully tokenized on
        *sub*tokens, use `is_pretokenized=False`.

        If *save* is ``True``, we'll keep the converted sentences as the
        Dataset source.

        If *append* is ``True``, we'll append the converted sentences to the
        existing Dataset source. Elsewise (default), the existing Dataset
        source will be replaced. The param is used only if *save* is
        ``True``."""

        tokenizer_max_len = self.tokenizer.max_len \
                                if hasattr(self.tokenizer, 'max_len') else \
                            self.tokenizer.model_max_len \
                                if hasattr(self.tokenizer,
                                           'model_max_len') else \
                            512
        if max_len == 0:
            max_len = tokenizer_max_len
        assert max_len and max_len >= 16, 'ERROR: max len must be >= 16'
        #assert max_len <= tokenizer_max_len, \
        #       'ERROR: max_len must be <= {}'.format(tokenizer_max_len)

        # TODO: not to cut output to 512 if max_len is None
        data = [
            self.tokenizer.encode_plus(
                text=sent,
                add_special_tokens=True,
                max_length=max_len,
                #is_pretokenized=is_pretokenized,
                pad_to_max_length=False,
                return_tensors=None,
                return_attention_mask=True,
                return_overflowing_tokens=False
            ) for sent in tqdm(
                iterable=sentences
                    if sentences and (isinstance(sentences[0], str)
                                   or is_pretokenized == False) else
                [' '.join(x for x in x) for x in sentences],
                mininterval=2, file=sys.stdout
            )
        ]
        if save:
            if append:
                self.data += data
            else:
                self.data = data
        else:
            return data

    def _collate(self, batch, with_lens=True):
        """The method to use with `torch.utils.data.DataLoader` and
        `.transform_collate()`.

        :with_lens: return lengths of data.
        :return: depends on keyword args.
        :rtype: tuple(
            dict({'input_ids': tensor([batch x batch max len]),
                  'token_type_ids': tensor([batch x batch max len]),
                  'attention_mask': tensor([batch x batch max len])}),
            lens:torch.tensor
        )
        """
        batch_ = {}
        [batch_.setdefault(k, []).append(v)
             for x in batch
             for k, v in x.items()]
        lens = [tensor([len(x['input_ids']) for x in batch],
                       dtype=self.int_tensor_dtype)] if with_lens else []
        return (self.tokenizer.pad(batch_, padding='longest',
                                   return_tensors='pt'),
                *lens)



# -*- coding: utf-8 -*-
# junky lib: dataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementagions of torch.utils.data.Dataset for different purposes.
"""
from junky.dataset.dummy_dataset import *
from junky.dataset.len_dataset import *
from junky.dataset.label_dataset import *

from junky.dataset.word_dataset import *
from junky.dataset.bert_dataset import *
from junky.dataset.bert_tokenized_dataset import *

from junky.dataset.char_dataset import *
from junky.dataset.token_dataset import *

from junky.dataset.frame_dataset import *
from junky.dataset.word_cat_dataset import *



# -*- coding: utf-8 -*-
# junky lib: layers.HighwayBiLSTM
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides a Highway LSTM implementagion for PyTorch models.
"""
import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence                               


class HBiLSTM_layer(nn.Module):
    """Highway LSTM implementation, modified from
    from https://github.com/bamtercelboo/pytorch_Highway_Networks/blob/master/models/model_HBiLSTM.py.
    [Original Article](https://arxiv.org/abs/1709.06436).

    Args:
        in_features:        number of features in input.
        out_features:       number of features in output.
        lstm_hidden_dim:    hidden dim for LSTM layer.
        lstm_num_layers:    number of LSTM layers.
        lstm_dropout:       dropout between 2+ LSTM layers.
        init_weight:        whether to init bilstm weights as xavier_uniform_
        init_weight_value:  bilstm weight initialization `gain` will be defined as 
            `np.sqrt(init_weight_value)`
    """

    def __init__(self, in_features, out_features,
                 lstm_hidden_dim, lstm_num_layers,
                 lstm_dropout=0.0, init_weight=True, 
                 init_weight_value=2.0):
        super().__init__()

        self.hidden_dim = lstm_hidden_dim
        self.num_layers = lstm_num_layers
        self.in_features = in_features
        self.out_features = out_features
        self.dropout = lstm_dropout
        self.bilstm = nn.LSTM(in_features, self.hidden_dim, num_layers=self.num_layers, bias=True, 
                              bidirectional=True, dropout=self.dropout)
        
        if init_weight:
            nn.init.xavier_uniform_(self.bilstm.all_weights[0][0], gain=np.sqrt(init_weight_value))
            nn.init.xavier_uniform_(self.bilstm.all_weights[0][1], gain=np.sqrt(init_weight_value))
            nn.init.xavier_uniform_(self.bilstm.all_weights[1][0], gain=np.sqrt(init_weight_value))
            nn.init.xavier_uniform_(self.bilstm.all_weights[1][1], gain=np.sqrt(init_weight_value))

        if self.bilstm.bias is True:
            a = np.sqrt(2/(1 + 600)) * np.sqrt(3)
            nn.init.uniform_(self.bilstm.all_weights[0][2], -a, a)
            nn.init.uniform_(self.bilstm.all_weights[0][3], -a, a)
            nn.init.uniform_(self.bilstm.all_weights[1][2], -a, a)
            nn.init.uniform_(self.bilstm.all_weights[1][3], -a, a)

        self.convert_layer = nn.Linear(in_features=self.hidden_dim * 2,
                                       out_features=self.in_features, bias=True)
        nn.init.constant_(self.convert_layer.bias, -1)

        self.fc1 = nn.Linear(in_features=self.in_features, out_features=self.hidden_dim*2, bias=True)
        nn.init.constant_(self.fc1.bias, -1)

        self.gate_layer = nn.Linear(in_features=self.in_features, out_features=self.hidden_dim*2, bias=True)
        nn.init.constant_(self.gate_layer.bias, -1)

    def forward(self, x, hidden, lens):
        # handle the source input x
        source_x = x
        
        x = pack_padded_sequence(x, lens, enforce_sorted=False)
        x, hidden = self.bilstm(x, hidden)
        x, _ = pad_packed_sequence(x)
        
        normal_fc = torch.transpose(x, 0, 1)

        x = x.permute(1, 2, 0)

        # normal layer in the formula is H
        source_x = torch.transpose(source_x, 0, 1)

        # the first way to convert 3D tensor to the Linear

        source_x = source_x.contiguous()
        information_source = source_x.view(source_x.size(0) * source_x.size(1), source_x.size(2))
        information_source = self.gate_layer(information_source)
        information_source = information_source.view(source_x.size(0), source_x.size(1), information_source.size(1))

        # transformation gate layer in the formula is T
        transformation_layer = torch.sigmoid(information_source)
        # carry gate layer in the formula is C
        carry_layer = 1 - transformation_layer
        # formula Y = H * T + x * C
        allow_transformation = torch.mul(normal_fc, transformation_layer)

        # the information_source compare to the source_x is for the same size of x,y,H,T
        allow_carry = torch.mul(information_source, carry_layer)
        # allow_carry = torch.mul(source_x, carry_layer)
        information_flow = torch.add(allow_transformation, allow_carry)

        information_flow = information_flow.contiguous()
        information_convert = information_flow.view(information_flow.size(0) * information_flow.size(1),
                                                    information_flow.size(2))
        information_convert = self.convert_layer(information_convert)
        information_convert = information_convert.view(information_flow.size(0), information_flow.size(1),
                                                       information_convert.size(1))

        information_convert = torch.transpose(information_convert, 0, 1)
        return information_convert, hidden


# HighWay recurrent model
class HighwayBiLSTM(nn.Module):
    """Highway LSTM model implementation, modified from
    from https://github.com/bamtercelboo/pytorch_Highway_Networks/blob/master/models/model_HBiLSTM.py.
    [Original Article](https://arxiv.org/abs/1709.06436).

    Args:
        hw_num_layers:      number of highway biLSTM layers.
        in_features:        number of features in input.
        out_features:       number of features in output.
        lstm_hidden_dim:    hidden dim for LSTM layer.
        lstm_num_layers:    number of LSTM layers.
        lstm_dropout:       dropout between 2+ LSTM layers.
        init_weight:        whether to init bilstm weights as xavier_uniform_
        init_weight_value:  bilstm weight initialization `gain` will be defined as 
            `np.sqrt(init_weight_value)`
        batch_first:        True if input.size(0) == batch_size.
        
    """

    def __init__(self, hw_num_layers, lstm_hidden_dim, lstm_num_layers, 
                 in_features, out_features, lstm_dropout,
                 init_weight=True, init_weight_value=2.0, batch_first=True):
        super().__init__()

        self.hw_num_layers = hw_num_layers
        self.hidden_dim = lstm_hidden_dim
        self.num_layers = lstm_num_layers
        self.batch_first = batch_first

        # multiple HighWay layers List
        self.highway = nn.ModuleList(
                        [HBiLSTM_layer(in_features, out_features, 
                                 lstm_hidden_dim, lstm_num_layers, 
                                 lstm_dropout=lstm_dropout, 
                                 init_weight=True, init_weight_value=2.0) 
                         for _ in range(hw_num_layers)]
        )
        self.output_layer = nn.Linear(in_features=in_features, out_features=out_features, bias=True)

        if self.output_layer.bias is True:
            a = np.sqrt(2/(1 + in_features)) * np.sqrt(3)
            nn.init.uniform_(self.output_layer, -a, a)

    def init_hidden(self, num_layers, batch_size, device):
        # the first is the hidden h
        # the second is the cell c
        return (Variable(torch.zeros(2 * num_layers, batch_size, self.hidden_dim)).to(device),
                Variable(torch.zeros(2 * num_layers, batch_size, self.hidden_dim)).to(device))

    def forward(self, x, lens):
        device = next(self.parameters()).device

        self.hidden = self.init_hidden(self.num_layers, x.size(0), device)

        if self.batch_first:
            x = x.transpose(0, 1)     # to (seq_len, batch_size, emb_dim)

        for current_layer in self.highway:
            x, self.hidden = current_layer(x, self.hidden, lens)

        x = torch.transpose(x, 0, 1)
        x = torch.tanh(x)
        output_layer = self.output_layer(x)

        return output_layer



# -*- coding: utf-8 -*-
# junky lib: dataset.WordDataset
#
# Copyright (C) 2020-present by Sergei Ternovykh, Anastasiya Nikiforova
# License: BSD, see LICENSE for details
"""
Provides implementation of torch.utils.data.Dataset for word-level input.
"""
from copy import deepcopy
#from gensim.models.keyedvectors import FastTextKeyedVectors
from junky import CPU, get_rand_vector, pad_sequences_with_tensor
from junky.dataset.base_dataset import BaseDataset
from torch import Tensor, float32, int64, tensor
from torch.nn.utils.rnn import pad_sequence


class WordDataset(BaseDataset):
    """
    `torch.utils.data.Dataset` for word-level input.

    Args:
        emb_model: dict or any other object that allow the syntax
            `vector = emb_model[word]` and `if word in emb_model:`.
        vec_size: the length of the word's vector.
        unk_token: add a token for words that are not present in the internal
            dict: str.
        unk_vec_norm: the norm of the vector for *unk_token*: float.
        pad_token: add a token for padding: str.
        pad_vec_norm: the norm of the vector for *pad_token*: float.
        extra_tokens: add tokens for any other purposes: list([str]).
        extra_vec_norm: the norm of the vectors for *extra_tokens*: float.
        float_tensor_dtype: dtype for float tensors: torch.dtype.
        int_tensor_dtype: dtype for int tensors: torch.dtype.
        sentences: sequences of words: list([list([str])]). If not ``None``,
            they will be transformed and saved.
        check_lower, skip_unk, keep_empty: params for the `.transform()`
        method.
    """
    #@property
    #def vec_size(self):
    #    return self.data[0].shape[-1] if self.data else \
    #           self.emb_model.vec_size
    #               if hasattr(self.emb_model, 'vec_size') else \
    #           self.emb_model.config.hidden_size

    def __init__(self, emb_model, vec_size,
                 unk_token=None, unk_vec_norm=1.,
                 pad_token=None, pad_vec_norm=0.,
                 extra_tokens=None, extra_vec_norm=1.,
                 float_tensor_dtype=float32, int_tensor_dtype=int64,
                 sentences=None, check_lower=True, skip_unk=False,
                 keep_empty=False):
        super().__init__()
        self.emb_model = emb_model
        self.vec_size = vec_size
        self.float_tensor_dtype = float_tensor_dtype
        self.int_tensor_dtype = int_tensor_dtype
        self.extra_model = {
            #t: emb_model[t]
            #       if t in (emb_model.vocab
            #                    if isinstance(emp_model,
            #                                  FastTextKeyedVectors) else
            #                emb_model) else
            #   get_rand_vector((vec_size,), extra_vec_norm)
            t: get_rand_vector((vec_size,), extra_vec_norm)
                for t in extra_tokens
        } if extra_tokens else \
        {}
        if unk_token:
            self.unk = self.extra_model[unk_token] = \
                get_rand_vector((vec_size,), unk_vec_norm)
        else:
            self.unk = None
        if pad_token:
            self.pad = self.extra_model[pad_token] = \
                get_rand_vector((vec_size,), pad_vec_norm)
            self.pad_tensor = tensor(self.pad, dtype=float_tensor_dtype)
        else:
            self.pad = None
            self.pad_tensor = None
        if sentences:
            self.transform(sentences, check_lower=check_lower,
                           skip_unk=skip_unk, keep_empty=keep_empty,
                           save=True)

    def _pull_xtrn(self):
        xtrn = self.emb_model
        self.emb_model = {}
        return xtrn

    def _push_xtrn(self, xtrn):
        self.emb_model = xtrn

    def word_to_vec(self, word, check_lower=True, force_lower=False,
                    skip_unk=True):
        """Convert a token to its vector. If the token is not present in the
        model, return vector of unk token or ``None`` if it's not defined.

        If *check_lower* is ``True`` (default), try to find vector of
        `word.lower()` if vector of *word* is absent.

        If *force_lower* is ``True``, convert *word* to lower case before
        looking its vectors up. Default is ``False``.

        If *skip_unk* is ``True``, unknown words will be skipped."""
        return self.extra_model[word] if word in self.extra_model else \
               self.emb_model[word.lower()] \
                   if force_lower and word.lower() in self.emb_model else \
               self.emb_model[word] if word in self.emb_model else \
               self.emb_model[word.lower()] \
                   if not force_lower \
                  and check_lower and word.lower() in self.emb_model else \
               self.unk if not skip_unk and self.unk is not None else \
               None

    def transform_words(self, words, check_lower=True, force_lower=False,
                        skip_unk=False):
        """Convert a word or a list of words to the corresponding
        vector|list of vectors.

        If *check_lower* is ``True`` (default), try to find a vector of a
        lower cased word if vector of a word is absent.

        If *force_lower* is ``True``, convert words to lower case before
        looking their vectors up. Default is ``False``.

        If *skip_unk* is ``True``, unknown words will be skipped."""
        return self.word_to_vec(words, check_lower=check_lower,
                                force_lower=force_lower, skip_unk=skip_unk) \
                   if isinstance(words, str) else \
               [self.word_to_vec(w, check_lower=check_lower,
                                 skip_unk=skip_unk) for w in words]

    def transform(self, sentences, check_lower=True, force_lower=False,
                  skip_unk=False, keep_empty=False, save=True, append=False):
        """Convert *sentences* of words to the sequences of the corresponding
        vectors and adjust their format for Dataset. If *skip_unk* is
        ``True``, unknown words will be skipped. If *keep_empty* is ``False``,
        we'll remove sentences that have no data after converting.

        If *check_lower* is ``True`` (default), try to find a vector of a
        lower cased word if vector of a word is absent.

        If *force_lower* is ``True``, convert sentences to lower case before
        looking their vectors up. Default is ``False``.

        If save is ``True``, we'll keep the converted sentences as the
        `Dataset` source.

        If *append* is ``True``, we'll append the converted sentences to the
        existing Dataset source. Elsewise (default), the existing Dataset
        source will be replaced. The param is used only if *save* is
        ``True``."""
        data = [tensor([
            v for v in s if keep_empty or v is not None
        ], dtype=self.float_tensor_dtype) for s in [
            self.transform_words(s, check_lower=check_lower,
                                 force_lower=force_lower, skip_unk=skip_unk) \
                for s in sentences
        ] if keep_empty or s]
        if save:
            if append:
                self.data += data
            else:
                self.data = data
        else:
            return data

    def _collate(self, batch, with_lens=True):
        """The method to use with `torch.utils.data.DataLoader` and
        `.transform_collate()`.

        :with_lens: return lengths of data.
        :return: depends on keyword args.
        :rtype: tuple(list([torch.tensor]), lens:torch.tensor)
        """
        assert self.pad_tensor is not None, \
               ('ERROR: pad_token must be defined if you want to use {} in '
                'DataLoader').format(self.__class__.__name__)
        device = batch[0].get_device() if batch[0].is_cuda else CPU
        lens = [tensor([len(x) for x in batch], device=device,
                       dtype=self.int_tensor_dtype)] if with_lens else []
        x = pad_sequences_with_tensor(batch, padding_tensor=self.pad_tensor)
        return (x, *lens) if lens else x
